package svc

import (
	"code.fnuoos.com/go_rely_warehouse/zyos_go_day_luck_draw.git/db"
	"code.fnuoos.com/go_rely_warehouse/zyos_go_day_luck_draw.git/db/model"
	"code.fnuoos.com/go_rely_warehouse/zyos_go_day_luck_draw.git/md"
	zhios_day_luck_draw_utils "code.fnuoos.com/go_rely_warehouse/zyos_go_day_luck_draw.git/utils"
	"code.fnuoos.com/go_rely_warehouse/zyos_go_day_luck_draw.git/utils/cache"
	"encoding/json"
	"errors"
	"fmt"
	"github.com/shopspring/decimal"
	"math/rand"
	"time"
	"xorm.io/xorm"
)

func InitForDayLuckDraw(redisAddr string) (err error) {
	if redisAddr != "" {
		cache.NewRedis(redisAddr)
	}
	_, err = cache.SelectDb(0)
	return
}

//随机几个用户
func RandUser(sess *xorm.Session) []string {
	var newOid = make([]string, 0)

	setting := db.GetDayLuckDrawSettingSess(sess)
	if setting == nil {
		return newOid
	}
	now := time.Now().Unix()
	if setting.IsUse == 0 || setting.StartTime.IsZero() || setting.EndTime.IsZero() {
		return newOid
	}
	if now < setting.StartTime.Unix() || now > setting.EndTime.Unix() {
		return newOid
	}
	count := 0
	var rewardData = make([]md.RewardData, 0)
	err := json.Unmarshal([]byte(setting.RewardData), &rewardData)
	if err == nil {
		for _, v := range rewardData {
			count += zhios_day_luck_draw_utils.StrToInt(v.Num)
		}
	}
	if count > 0 {
		oids := db.GetAllOrderOid(sess)
		newOid = MicsSlice(oids, count)
		return newOid
	}
	return newOid

}

/***
1.创建期数
2.获取还没分配的金额 就是 期数为0的
3.把期数设置到订单
4.查出所有订单号
5.随机出订单号（每个名额随机一次 如果重复再次循环） 按概率分几等奖
*/

func GetUserReward(eg *xorm.Engine, mid string, isTask bool) error {
	// 加锁 防止频繁操作
	mutexKey := fmt.Sprintf("%s:day_lucky_raw_settle", mid)
	operationSumAvailable, err := cache.Do("SET", mutexKey, 1, "EX", 3600, "NX")
	if err != nil {
		return err
	}
	defer cache.Del(mutexKey)
	if operationSumAvailable != "OK" {
		return errors.New("上一次未运行完成")
	}
	sess := eg.NewSession()
	defer func() {
		sess.Close()
		if err := recover(); err != nil {
			fmt.Println(err)
		}
	}()
	sess.Begin()
	key := fmt.Sprintf("%s:day_lucky_draw_bingo_rate_calculate", mid)
	isDone, _ := cache.GetString(key)
	if isDone == "" || isDone == "0" {
		SetAllAwardsProbability(sess, mid)
	}
	probabilityMap := make([]map[int]int64, 0)
	key = fmt.Sprintf("%s:day_lucky_draw_rotary_bingo_rate", mid)
	err = cache.GetJson(key, &probabilityMap)
	if err != nil {
		sess.Rollback()
		return err
	}

	setting := db.GetDayLuckDrawSettingSess(sess)
	var rewardData = make([]md.RewardData, 0)
	err = json.Unmarshal([]byte(setting.RewardData), &rewardData)
	if err != nil {
		sess.Rollback()
		return err
	}
	now := time.Now().Unix()
	if isTask {
		if setting.IsUse == 0 || setting.StartTime.IsZero() || setting.EndTime.IsZero() {
			return errors.New("未开始")
		}
		if now < setting.StartTime.Unix() || now > setting.EndTime.Unix() {
			return errors.New("未开始")
		}
		if setting.OpenTime.IsZero() {
			return errors.New("没抽奖时间")
		}
		if now < setting.OpenTime.Unix() {
			return errors.New("没到时间")
		}
	}
	if setting.Source == "" {
		return errors.New("未选择奖励池数据类型")
	}
	//设置期数
	max := db.GetPeriodMax(sess)
	sum := db.GetSettleAllOrderSum(sess)
	if setting.Source == "amount" {
		sum = zhios_day_luck_draw_utils.StrToFloat64(setting.Amount)
	} else {
		sum = sum * zhios_day_luck_draw_utils.StrToFloat64(setting.CommissionBili) / 100
	}
	if sum == 0 { //没金额中断了
		setting.OpenTime = time.Unix(now+int64(setting.CycleDay)*86400, 0)
		update, err := sess.Where("id=?", setting.Id).Cols("open_time").Update(setting)
		if err != nil || update == 0 {
			sess.Rollback()
			return errors.New("失败")
		}
		sess.Commit()
		return errors.New("没金额")
	}
	period := 1
	if max != nil {
		period = max.Period + 1
	}
	req := md.DayLuckDrawPeriodReq{
		Reward: zhios_day_luck_draw_utils.Float64ToStrByPrec(sum, 7),
		Period: zhios_day_luck_draw_utils.IntToStr(period),
	}
	settingStr, _ := json.Marshal(setting)
	bools := db.AddPeriod(sess, req, string(settingStr))
	if bools == false {
		return errors.New("期数写入错误")
	}

	rewardDataMap := make(map[int64]int, 0)
	for _, v := range rewardData {
		rewardDataMap[zhios_day_luck_draw_utils.StrToInt64(v.Key)] = zhios_day_luck_draw_utils.StrToInt(v.Num)
	}
	uids := RandUser(sess)
	var award = make(map[string]int64)
	var awardIdMap = make(map[int64]int)
	for _, v := range uids {
		for i := 0; i < 10; i++ {
			key := AwardRandomNumberFor100000()
			awardId := int64(0)
			for _, awardProbabilityMap := range probabilityMap {
				id, ok := awardProbabilityMap[key]
				if ok {
					awardId = id
				} else {
					continue
				}
			}
			if awardId == 0 || award[v] > 0 {
				continue
			}
			//查下已产生用户数量
			awardCount, ok := awardIdMap[awardId]
			//查下总数量
			rewardDataCount, ok := rewardDataMap[awardId]
			if ok == false {
				continue
			}
			if ok {
				if awardCount+1 > rewardDataCount {
					continue
				}
			}
			if rewardDataCount == 0 {
				continue
			}
			awardIdMap[awardId] = awardCount + 1
			award[v] = awardId
		}
	}

	//处理奖励
	var awardMap = make(map[int64][]string)
	for k, v := range award {
		awardMap[v] = append(awardMap[v], k)
	}
	for _, v := range rewardData {
		uid, ok := awardMap[zhios_day_luck_draw_utils.StrToInt64(v.Key)]
		if ok {
			err := SendReward(sess, uid, mid, period, sum, v)
			if err != nil {
				sess.Rollback()
				return err
			}
		}
	}
	setting.OpenTime = time.Unix(now+int64(setting.CycleDay)*86400, 0)
	update, err := sess.Where("id=?", setting.Id).Cols("open_time").Update(setting)
	if err != nil || update == 0 {
		sess.Rollback()
		return errors.New("失败")
	}
	update1, err1 := sess.Where("is_send=0").Cols("is_send").Update(&model.DayLuckDrawSettleOrder{IsSend: 1})
	if err1 != nil || update1 == 0 {
		sess.Rollback()
		return errors.New("失败")
	}
	update2, err2 := sess.Where("period=0").Cols("period").Update(&model.DayLuckDrawOrderJackpot{Period: period})
	if err2 != nil || update2 == 0 {
		sess.Rollback()
		return errors.New("失败")
	}
	sess.Commit()
	return nil
}
func SendReward(sess *xorm.Session, uid []string, mid string, period int, sum float64, rewardData md.RewardData) error {
	count := len(uid)
	if count == 0 {
		return nil
	}
	bili := zhios_day_luck_draw_utils.StrToFloat64(rewardData.Bili)
	if bili == 0 {
		return nil
	}
	amount := sum * bili / float64(count)
	for _, v := range uid {
		var req = md.DayLuckDrawRewardReq{
			Uid:    v,
			Period: zhios_day_luck_draw_utils.IntToStr(period),
			Reward: zhios_day_luck_draw_utils.Float64ToStrByPrec(amount, 7),
			Lv:     rewardData.Key,
		}
		err := db.AddPeriodReward(sess, req)
		if err != nil {
			return err
		}
		var dealUserAmount = md.DealUserAmount{
			Kind:      "add",
			Mid:       mid,
			Title:     "每日抽奖奖励",
			OrderType: "",
			OrdAction: 85,
			OrdId:     "",
			Uid:       zhios_day_luck_draw_utils.StrToInt(v),
			Amount:    amount,
		}
		err = DealUserAmount(sess, dealUserAmount)
		if err != nil {
			return err
		}
	}
	return nil
}
func AwardRandomNumberFor100000() int {
	rand.Seed(time.Now().UnixNano())
	num := rand.Intn(100000)
	return num
}

func MicsSlice(origin []string, count int) []string {
	tmpOrigin := make([]string, len(origin))
	copy(tmpOrigin, origin)
	//一定要seed
	rand.Seed(time.Now().Unix())
	rand.Shuffle(len(tmpOrigin), func(i int, j int) {
		tmpOrigin[i], tmpOrigin[j] = tmpOrigin[j], tmpOrigin[i]
	})

	result := make([]string, 0)
	for _, value := range tmpOrigin {
		if len(result) == count {
			break
		}
		if zhios_day_luck_draw_utils.InArr(value, result) {
			continue
		}
		result = append(result, value)
	}
	return result
}

func SetAllAwardsProbability(sess *xorm.Session, mid string) error {
	defer func() {
		//处理异常错误
		if r := recover(); r != nil {
			fmt.Println(r)
		}
	}()
	setting := db.GetDayLuckDrawSettingSess(sess)
	var rewardData = make([]md.RewardData, 0)
	err := json.Unmarshal([]byte(setting.RewardData), &rewardData)
	if err != nil {
		key := fmt.Sprintf("%s:day_lucky_draw_bingo_rate_calculate", mid)
		_, err = cache.SetEx(key, "0", 39528000) // 半年
		if err != nil {
			return err
		}
	}
	//建立协程池
	workerCount := 0
	ProbabilityKeyChan := make(chan int, 100000)
	awardsMapChan := make(chan map[int]int64, len(rewardData))
	doneChan := make(chan bool, len(rewardData))
	for i := 0; i < 10; i++ {
		go func(endNum int) {
			for startNum := (endNum * 10000) + 1; startNum <= (endNum+1)*10000; startNum++ {
				ProbabilityKeyChan <- startNum
			}
			if i == 9 {
				close(ProbabilityKeyChan)
			}
		}(i)
	}
	for _, award := range rewardData {
		probabilityD, err := decimal.NewFromString(award.Num)
		if err != nil {
			probabilityD = decimal.NewFromInt(0)
		}
		probabilityD = probabilityD.Div(decimal.NewFromInt(100))
		probability, _ := probabilityD.Float64()
		go CalculateAwardsProbability(zhios_day_luck_draw_utils.StrToInt64(award.Key), probability, ProbabilityKeyChan, awardsMapChan, doneChan)
		workerCount++
	}
	probabilityMap := make([]map[int]int64, 0)
	WaitChain(workerCount, awardsMapChan, doneChan)
	for awards := range awardsMapChan {
		probabilityMap = append(probabilityMap, awards)
	}
	key := fmt.Sprintf("%s:day_lucky_draw_rotary_bingo_rate", mid)
	_, err = cache.SetEx(key, zhios_day_luck_draw_utils.SerializeStr(probabilityMap), 39528000) // 半年
	if err != nil {
		return err
	}
	key = fmt.Sprintf("%s:day_lucky_draw_bingo_rate_calculate", mid)
	_, err = cache.SetEx(key, "1", 39528000) // 半年
	if err != nil {
		return err
	}
	return nil
}

func CalculateAwardsProbability(awardsId int64, probability float64, ProbabilityKeyChan chan int, awardsMapChan chan map[int]int64, doneChan chan bool) {
	//计算要循环多少次
	var awardsMap = make(map[int]int64)
	for i := 0; i < int(100000*probability); i++ {
		key := <-ProbabilityKeyChan
		awardsMap[key] = awardsId
	}
	awardsMapChan <- awardsMap
	doneChan <- true
}

func WaitChain(workerCount int, awardsMapChan chan map[int]int64, doneChan chan bool) {
	for {
		select {
		case <-doneChan:
			workerCount--
			if workerCount == 0 {
				close(awardsMapChan)
				return
			}
		}
	}
}
